import random
import sqlite3
from contextlib import contextmanager
from pathlib import Path


from ardilla import Engine, Model, Crud
from ardilla.errors import QueryExecutionError
from pydantic import Field

from ardilla.fields import ForeignField


path = Path(__file__).parent
db = path / "testdb_sync.sqlite"


class User(Model):
    id: int = Field(primary=True, autoincrement=True)
    name: str
    age: int


@contextmanager
def cleanup():
    try:
        db.unlink(missing_ok=True)
        engine = Engine(db, single_connection=True)
        crud = engine.crud(User)
        yield crud
    finally:
        engine.close()
        db.unlink(missing_ok=True)


@contextmanager
def query():
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        yield cur
    finally:
        cur.close()
        con.close()


"""
Abstract methods:
    insert x
    insert_or_ignore x
    save_one x
    save_many x
    get_all x
    get_many x
    get_or_create x
    get_or_none x 
    delete_one x
    delete_many x
"""


def test_insert():
    with cleanup() as crud:
        u = crud.insert(id=1, name="chris", age=35)
        assert u is not None, "User wasn't created as expected"
        assert u.__rowid__ is not None, "Created user did not have __rowid__ set"
        assert u.__rowid__ == 1, "Created User did not have correct __rowid__ "
        try:
            crud.insert(id=1, name="chris", age=35)
        except QueryExecutionError:
            pass
        else:
            raise Exception("QueryExcecutionError should have been rised")


def test_insert_or_ignore():
    with cleanup() as crud:
        u = crud.insert_or_ignore(id=1, name="chris", age=35)
        assert u is not None, "User wasn't created as expected"
        u = crud.insert_or_ignore(id=1, name="chris", age=35)
        assert u is None, "User was returned but it should have been None"


def test_save_one():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        crud.save_one(chris)

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 1, "Incorrect number of users in the database"


def test_save_one_rowid():
    with cleanup() as crud:
        chris = crud.insert(id=1, name="chris", age=35)
        chris.name = 'chris_is_alt'
        crud.save_one(chris)
        

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 1, "Incorrect number of users in the database"

def test_save_many():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, moni, elena)

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 3, "Incorrect number of users in the database"


def test_get_all():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, moni, elena)

        users = crud.get_all()
        assert len(users) == 3, "Incorrect number of users returned"


def test_get_many():
    with cleanup() as crud:
        u1 = User(id=1, name="chris", age=35)
        u2 = User(id=2, name="chris", age=35)
        u3 = User(id=3, name="chris", age=35)
        u4 = User(id=4, name="moni", age=34)
        u5 = User(id=5, name="fran", age=1)
        crud.save_many(u1, u2, u3, u4, u5)

        users = crud.get_many(name="chris")

        assert len(users) == 3, "Incorrect number of users returned"
        assert all(u.__rowid__ for u in users), 'User objects did not get their rowid populated'

def test_get_many_with_limit():
    users = [User(id=n, name='chris', age=random.randint(1, 27)) for n in range(50)]
    with cleanup() as crud:
        crud.save_many(*users)
        users = crud.get_many(limit=5)
        assert len(users) == 5

def test_get_many_with_ordering():
    with cleanup() as crud:
        crud.insert(id=1, name='chris', age=3)
        crud.insert(id=2, name='chris', age=2)
        crud.insert(id=3, name='chris', age=5)
        crud.insert(id=4, name='chris', age=1)
        crud.insert(id=5, name='chris', age=4)
        
        users = crud.get_many(order_by={'age': 'desc'})
        
        sorted_users = sorted(users, key=lambda u: u.age, reverse=True)
        assert users == sorted_users


def test_get_or_create():
    with cleanup() as crud:
        elena = User(id=1, name="elena", age=5)
        crud.save_one(elena)

        u, c = crud.get_or_create(name="fran", age=1)
        assert c is True, "User should have been created"
        assert u == User(id=2, name="fran", age=1), f"Unexpected user was created: {u}"
        u, c = crud.get_or_create(id=1)
        assert u == elena, "Mismatch between expected and returned user"
        assert c is False, "User shouldn't have been created"


def test_get_or_none():
    with cleanup() as crud:
        elena = User(id=1, name="elena", age=5)
        crud.save_one(elena)

        u = crud.get_or_none(id=1)
        assert u is not None, "User wasn't found but it should have"
        u = crud.get_or_none(id=2)
        assert u is None, "User shouln't have been found but was"


def test_delete_one():
    with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        crud.save_many(chris, moni, elena)
        crud.delete_one(moni)
        users = crud.get_all()
        assert len(users) == 2, "Delete one didn't delete the correct amount of users"

class Foo(Model):
    a: str
    b: int

def test_delete_one_without_ids():
    db.unlink(missing_ok=True)
    engine = Engine(db)
    crud = engine.crud(Foo)
    try:
        for i, l in enumerate('abcdef',1):
            crud.insert(a=l, b=i)
            
        to_del = [
            Foo(a='b', b=2),
            Foo(a='c', b=3),
            Foo(a='d', b=4),
        ]
        for obj in to_del:
            crud.delete_one(obj)
        
        foos = crud.get_all()
        assert len(foos) == 3, 'Mismatch beween expected and found after delete_many'
        
    finally:
        db.unlink(missing_ok=True)
    

def test_delete_many_by_id():
    with cleanup() as crud:
        users = [
            User(id=n, name='chris', age=n)
            for n in range(10)
        ]
        crud.save_many(*users)
        
        to_delete = users[:-1]
        crud.delete_many(*to_delete)
        
        users = crud.get_all()
        assert len(users) == 1, "Delete many didn't delete the correct amount of users"


def test_delete_many_by_rowid():
    with cleanup() as crud:
        users = [
            User(id=n, name='chris', age=n)
            for n in range(10)
        ]
        crud.save_many(*users)
       
        crud.delete_many(*users[:-1])
        
        users = crud.get_all()
        
        
        assert len(users) == 1, "Delete many didn't delete the correct amount of users"


def test_foreign_keys():
    db = path / 'sync_test.sqlite'
    db.unlink(missing_ok=True)
    engine = Engine(db, enable_foreing_keys=True)
    
    class Guild(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        
    class User(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        guild_id: int = ForeignField(references=Guild, on_delete=ForeignField.CASCADE)
    
    gcrud = engine.crud(Guild)
    ucrud = engine.crud(User)
    
    ga = gcrud.insert(name='guild a')
    gb = gcrud.insert(name='guild b')
    for guild in [ga, gb]:
        for n in range(5):
            ucrud.insert(name=f'user {n}', guild_id=guild.id)
    
    users = ucrud.get_all()
    assert len(users) == 10
    gcrud.delete_one(ga)
    users = ucrud.get_all()
    assert len(users) == 5 
    db.unlink(missing_ok=True)


from abc import ABC, abstractmethod
import aiosqlite

class AbstractAsyncEngine(ABC):
    """This just provides autocompletition across the library"""
    schemas: set[str]

    @abstractmethod
    async def __aenter__(self) -> aiosqlite.Connection:
        ...
    @abstractmethod
    async def cursor(self, con: aiosqlite.Connection) -> aiosqlite.Cursor:
        ...
    @abstractmethod
    async def connect(self) -> aiosqlite.Connection:
        ...


from __future__ import annotations
import aiosqlite

from .crud import AsyncCrud
from ..models import M
from ..engine import Engine

from .abc import AbstractAsyncEngine

class AsyncEngine(Engine, AbstractAsyncEngine):
    def __init__(self, path: str, enable_foreing_keys: bool = False, single_connection: bool = False):
        if single_connection is True:
            raise NotImplementedError('The async engine does not support single connection for now')
        super().__init__(path, enable_foreing_keys, single_connection)
    
    async def connect(self) -> aiosqlite.Connection:
        con = await aiosqlite.connect(self.path)
        con.row_factory = aiosqlite.Row
        return con

    async def __aenter__(self) -> aiosqlite.Connection:
        self.acon = await self.connect()
        if self.enable_foreing_keys:
            await self.acon.execute('PRAGMA foreign_keys = ON;')
            
        return self.acon

    async def __aexit__(self, *_):
        await self.acon.close()

    async def setup(self):
        async with self as con:
            for table in self.schemas:
                await con.execute(table)
                self.tables_created.add(table)
            await con.commit()
    
    def crud(self, Model: type[M]) -> AsyncCrud[M]:
        """This function runs synchronously"""
        crud = self._cruds.setdefault(Model, AsyncCrud(Model, self))
        if Model.__schema__ not in self.tables_created:
            with self as con:
                con.execute(Model.__schema__)
                con.commit()
            self.tables_created.add(Model.__schema__)
        return crud


"""
Contains the module's errors
"""

class ArdillaException(Exception):
    pass


class ModelIntegrityError(ArdillaException):
    pass


class MissingEngine(ArdillaException):
    pass


class QueryExecutionError(ArdillaException):
    pass


class BadQueryError(ArdillaException):
    pass



import sqlite3
from sqlite3 import Row
from typing import Literal, Generic, Optional, Union

from .abc import CrudABC, AbstractEngine
from .models import M
from .errors import QueryExecutionError
from .logging import log
from . import queries


class Crud(CrudABC, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""

    engine: AbstractEngine

    def _do_insert(
        self,
        ignore: bool = False,
        returning: bool = True,
        /,
        **kws,
    ) -> Optional[M]:
        q, vals = queries.for_do_insert(self.tablename, ignore, returning, kws)

        with self.engine as con:
            with self.engine.cursor(con) as cur:
                try:
                    cur.execute(q, vals)
                except sqlite3.IntegrityError as e:
                    raise QueryExecutionError(str(e))

                row = cur.fetchone()
                con.commit()
                if returning and row:
                    return self._row2obj(row, cur.lastrowid)

        return None

    def get_or_none(self, **kws) -> Optional[M]:
        """Gets an object from a database or None if not found"""
        q, vals = queries.for_get_or_none(self.tablename, kws)
        with self.engine as con:
            ctxcur = self.engine.cursor(con)
            with ctxcur as cur:
                cur.execute(q, vals)
                row: Union[Row, None] = cur.fetchone()
                if row:
                    return self._row2obj(row)
        return None

    def insert(self, **kws) -> M:
        """
        Inserts a record into the database.
        Returns:
            Model: Creates a new entry in the database and returns the object
        Rises:
            ardilla.error.QueryExecutionError: if there's a conflict when inserting the record
        """
        return self._do_insert(False, True, **kws)

    def insert_or_ignore(self, **kws) -> Optional[M]:
        """inserts a the object of a row or ignores it if it already exists"""
        return self._do_insert(True, True, **kws)

    def get_or_create(self, **kws) -> tuple[M, bool]:
        """Returns object and bool indicated if it was created or not"""
        created = False
        result = self.get_or_none(**kws)
        if not result:
            result = self.insert_or_ignore(**kws)
            created = True
        return result, created

    def get_all(self) -> list[M]:
        """Gets all objects from the database"""
        return self.get_many()

    def get_many(
        self,
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        **kws,
    ) -> list[M]:
        """Returns a list of objects that have the given conditions"""
        q, vals = queries.for_get_many(self.Model, order_by=order_by, limit=limit, kws=kws)
        with self.engine as con:
            ctxcur = self.engine.cursor(con)
            with ctxcur as cur:
                cur.execute(q, vals)
                rows: list[Row] = cur.fetchall()
                return [self._row2obj(row) for row in rows]

    def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database"""
        q, vals = queries.for_save_one(obj)

        with self.engine as con:
            con.execute(q, vals)
            con.commit()
        return True

    def save_many(self, *objs: tuple[M]) -> Literal[True]:
        """Saves all the given objects to the database"""
        q, vals = queries.for_save_many(objs)
        with self.engine as con:
            con.executemany(q, vals)
            con.commit()

        return True

    def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        If the object has a PK field or the rowid setup, those will be
        used to locate the obj and delete it.
        If not, this function will delete any object like this one.
        """

        q, vals = queries.for_delete_one(obj)
        with self.engine as con:
            con.execute(q, vals)
            con.commit()

        return True

    def delete_many(self, *objs: M) -> Literal[True]:
        q, vals = queries.for_delete_many(objs)
        with self.engine as con:
            con.execute(q, vals)
            con.commit()

        return True



from typing import Literal, Generic, Optional, Union

import aiosqlite
from aiosqlite import Row

from ..errors import QueryExecutionError
from ..models import M
from ..abc import CrudABC
from ..logging import log
from .. import queries

from .abc import AbstractAsyncEngine


class AsyncCrud(CrudABC, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""

    engine: AbstractAsyncEngine

    async def get_or_none(self, **kws) -> Optional[M]:
        """Gets an object from a database or None if not found"""
        q, vals = queries.for_get_or_none(self.tablename, kws)
        async with self.engine as con:
            async with con.execute(q, vals) as cur:
                row: Union[Row, None] = await cur.fetchone()
                if row:
                    return self._row2obj(row)
        return None

    async def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        q, vals = queries.for_do_insert(self.tablename, ignore, returning, kws)

        async with self.engine as con:
            con = await self.engine.connect()
            cur = None
            try:
                cur = await con.execute(q, vals)
            except aiosqlite.IntegrityError as e:
                raise QueryExecutionError(str(e))
            else:
                row = await cur.fetchone()
                await con.commit()
                if returning and row:
                    return self._row2obj(row, cur.lastrowid)
            finally:
                if cur is not None:
                    await cur.close()
                await con.close()

    async def insert(self, **kws):
        """
        Inserts a record into the database.
        Returns:
            Model | None: Returns a model only if newly created
        Rises:
            ardilla.error.QueryExecutionError: if there's a conflict when inserting the record
        """
        return await self._do_insert(False, True, **kws)

    async def insert_or_ignore(self, **kws) -> Optional[M]:
        """inserts a the object of a row or ignores it if it already exists"""
        return await self._do_insert(True, True, **kws)

    async def get_or_create(self, **kws) -> tuple[M, bool]:
        """Returns object and bool indicated if it was created or not"""
        created = False
        result = await self.get_or_none(**kws)
        if not result:
            result = await self.insert_or_ignore(**kws)
            created = True
        return result, created

    async def get_all(self) -> list[M]:
        """Gets all objects from the database"""
        q = f"SELECT rowid, * FROM {self.tablename};"
        log.debug(f"Querying: {q}")

        async with self.engine as con:
            async with con.execute(q) as cur:
                return [self._row2obj(row) for row in await cur.fetchall()]

    async def get_many(
        self,
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        **kws,
    ) -> list[M]:
        """Returns a list of objects that have the given conditions"""
        q, vals = queries.for_get_many(self.Model, order_by=order_by, limit=limit, kws=kws)
        async with self.engine as con:
            async with con.execute(q, vals) as cur:
                rows: list[Row] = await cur.fetchall()
                return [self._row2obj(row) for row in rows]

    async def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database"""
        q, vals = queries.for_save_one(obj)

        async with self.engine as con:
            await con.execute(q, vals)
            await con.commit()
        return True

    async def save_many(self, *objs: M) -> Literal[True]:
        """Saves all the given objects to the database"""
        q, vals = queries.for_save_many(objs)

        async with self.engine as con:
            await con.executemany(q, vals)
            await con.commit()

        return True

    async def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        queries only by the Model id fields (fields suffixed with 'id')
        """
        q, vals = queries.for_delete_one(obj)

        async with self.engine as con:
            await con.execute(q, vals)
            await con.commit()
        return True

    async def delete_many(self, *objs: M) -> Literal[True]:
        q, vals = queries.for_delete_many(objs)

        async with self.engine as con:
            await con.execute(q, vals)
            await con.commit()

        


from typing import Container

def validate_ordering(columns: Container[str], order_by: dict[str, str]) -> dict[str, str]:
    """validates an ordering dictionary
    The ordering should have this structure:
        {
            'column_name': 'ASC' OR 'DESC'
        }
    Case in values is insensitive
    
    Args:
        columns (Container[str]): a collection of columns to check the keys against
        order_by (dict[str, str]): 

    Raises:
        KeyError: if the key is not listed in the columns of the table
        ValueError: if the value is not either ASC or DESC

    Returns:
        dict[str, str]: a copy of the ordering dict with values in uppercase
    """
    out = order_by.copy()
    for k, v in order_by.items():
        if k not in columns:
            raise KeyError(f'"{k}" is not a valid column name')
        elif v.lower() not in {'desc', 'asc'}:
            raise ValueError(f'"{k}" value "{v}" is invalid, must be either "asc" or "desc" (case insensitive)')
        else:
            out[k] = v.upper()
    return out


"""
Contains the Model object and typing alias to work with the engines and Cruds
"""

from typing import Optional, TypeVar
from pydantic import BaseModel, PrivateAttr

from .schemas import make_table_schema, FIELD_MAPPING, get_tablename, get_pk
from .errors import ModelIntegrityError


class Model(BaseModel):
    """
    The base model representing SQLite tables
    Inherits directly from pydantic.BaseModel
    
    Additional Attributes:
        __rowid__ (int | None): when an object is returned by a query it will 
            contain the rowid field that can be used for update and deletion.
        __pk__ (str | None): Holds the primary key column name of the table
        __tablename__ (str): the name of the table in the database
        __schema__(str): the schema for the table. Should start with `CREATE TABLE IF NOT EXISTS`
        
    Usage Example:
        from ardilla import Model, Field
        # Field is actually pydantic.Field but it's imported here for the convenience of the developer
        
        class User(Model):
            __tablename__ = 'users' # by default the tablename is just the model's name in lowercase
            id: int = Field(primary=True) # sets this field as the primary key
            name: str
    """
    __rowid__: Optional[int] = PrivateAttr(default=None)
    __pk__: Optional[str]  # tells the model which key to idenfity as primary
    __tablename__: str  # will default to the lowercase name of the subclass
    __schema__: str  # best effort will be made if it's missing
    # there's no support for constrains or foreign fields yet but you can
    # define your own schema to support them

    def __init_subclass__(cls, **kws) -> None:
        pk = None
        for field in cls.__fields__.values():
            if field.type_ not in FIELD_MAPPING:
                raise ModelIntegrityError(
                    f'Field "{field.name}" of model "{cls.__name__}" is of unsupported type "{field.type_}"'
                )
            if (field.field_info.extra.get('primary') 
                or field.field_info.extra.get('primary_key')):
                if getattr(cls, '__pk__', None) is not None:
                    raise ModelIntegrityError('More than one fields defined as primary')
                
                cls.__pk__ = field.name 

        if not hasattr(cls, "__schema__"):
            cls.__schema__ = make_table_schema(cls)
        
        if not hasattr(cls, '__pk__'):
            cls.__pk__ = get_pk(cls.__schema__)

        if not hasattr(cls, "__tablename__"):
            tablename = get_tablename(cls)
            setattr(cls, "__tablename__", tablename)

        super().__init_subclass__(**kws)

    def __str__(self) -> str:
        return f"{self!r}"


M = TypeVar("M", bound=Model)



import sqlite3
from typing import Self, Literal, TypeVar, Protocol, Optional
from abc import abstractmethod, ABC
from sqlite3 import Row

from .errors import MissingEngine
from .models import M, Model as BaseModel

E = TypeVar("E")  # Engine Type


class ContextCursorProtocol(Protocol):
    def __init__(self, con: sqlite3.Connection):
        ...

    def __enter__(self) -> sqlite3.Cursor:
        ...

    def __exit__(self, *_) -> None:
        ...


class AbstractEngine(ABC):
    """This just provides autocompletition across the library"""

    schemas: set[str]

    @abstractmethod
    def __enter__(self) -> sqlite3.Connection:
        ...

    @abstractmethod
    def __exit__(self, *_) -> None:
        ...

    @abstractmethod
    def cursor(self, con: sqlite3.Connection) -> ContextCursorProtocol:
        ...


class CrudABC(ABC):
    __slots__ = (
        "engine",
        "tablename",
        "Model",
        "columns",
    )

    engine: AbstractEngine

    def __init__(self, Model: type[M], engine: Optional[AbstractEngine] = None) -> None:
        if engine:
            self.engine = engine

        self.Model = Model
        if Model.__schema__:
            self.engine.schemas.add(Model.__schema__)
        self.tablename = Model.__tablename__
        self.columns = tuple(Model.__fields__)

    def __new__(cls, Model: type[M], engine: Optional[AbstractEngine] = None) -> Self:
        if not issubclass(Model, BaseModel):
            raise TypeError("Model param has to be a subclass of model")

        cls_engine = getattr(cls, "engine", None)

        if engine is None and cls_engine is None:
            # if not isinstance(engine, Engine) and not isinstance(cls_engine, Engine):
            raise MissingEngine(
                "Missing engine. Set the engine at instance level (Crud(Model, engine))"
                "or at class level (Crud.engine = engine)"
            )
        return super().__new__(cls)

    def _row2obj(self, row: Row, rowid: Optional[int] = None) -> BaseModel:
        """
        Args:
            row: the sqlite row
            rowid: the rowid of the row.
                If passed it means it comes from an insert function

        """
        [*keys] = self.Model.__fields__
        if rowid is None:
            rowid, *vals = row
        else:
            vals = list(row)
        data = {k: v for k, v in zip(keys, vals)}

        obj = self.Model(**data)
        obj.__rowid__ = rowid
        return obj

    # Create
    @abstractmethod
    def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        pass

    @abstractmethod
    def insert(self, **kws):
        pass

    @abstractmethod
    def insert_or_ignore(self):
        pass

    # Read
    @abstractmethod
    def get_all(self) -> list[M]:
        pass

    @abstractmethod
    def get_many(self, order_by: Optional[dict[str, str]] = None, limit: Optional[int] = None, **kws) -> list[M]:
        pass

    @abstractmethod
    def get_or_create(self, **kws) -> tuple[M, bool]:
        pass

    @abstractmethod
    def get_or_none(self, **kws) -> Optional[M]:
        pass

    # Update
    @abstractmethod
    def save_one(self, obj: M) -> Literal[True]:
        pass

    @abstractmethod
    def save_many(self, *objs: M) -> Literal[True]:
        pass

    # Delete
    @abstractmethod
    def delete_one(self, obj: M) -> Literal[True]:
        pass

    @abstractmethod
    def delete_many(self, *objs: M) -> Literal[True]:
        pass



from __future__ import annotations
import sqlite3

from .models import M
from .crud import Crud
from .abc import AbstractEngine, ContextCursorProtocol
from .logging import log


class ContextCursor(ContextCursorProtocol):
    __slots__ = ('con', 'cur')
    def __init__(self, con: sqlite3.Connection):
        self.con = con

    def __enter__(self) -> sqlite3.Cursor:
        self.cur = self.con.cursor()
        return self.cur

    def __exit__(self, *_) -> None:
        self.cur.close()


class Engine(AbstractEngine):
    __slots__ = (
        'path',
        'schemas',
        '_cruds',
        'tables_created',
        'enable_foreing_keys',
        'con',
        'single_connection',
        'acon',
    )
    def __init__(
        self,
        path: str,
        enable_foreing_keys: bool = False,
        single_connection: bool = False,
    ):
        self.path = path
        self.schemas: set[str] = set()
        self._cruds: dict[type[M], Crud[M]] = {}
        self.tables_created: set[str] = set()
        self.enable_foreing_keys = enable_foreing_keys
        self.single_connection = single_connection
        log.debug(f"Instantiating {self.__class__.__name__}")

    def __enter__(self) -> sqlite3.Connection:
        if not hasattr(self, 'con'):    
            con = sqlite3.connect(self.path)
            con.row_factory = sqlite3.Row
            self.con = con
            if self.enable_foreing_keys:
                self.con.execute("PRAGMA foreign_keys = on;")

        return self.con

    def __exit__(self, *_):
        if not self.single_connection:
            self.con.close()
            delattr(self, 'con')
    
    def close(self):
        if hasattr(self, 'con'):
            self.con.close()
            delattr(self, 'con')
        
            

    def cursor(self, con: sqlite3.Connection) -> ContextCursor:
        return ContextCursor(con)

    def setup(self):
        with self as con:
            for table in self.schemas:
                con.execute(table)
                self.tables_created.add(table)
            con.commit()

    def crud(self, Model: type[M]) -> Crud[M]:
        crud = self._cruds.setdefault(Model, Crud(Model, self))
        if Model.__schema__ not in self.tables_created:
            with self as con:
                con.execute(Model.__schema__)
                con.commit()
            self.tables_created.add(Model.__schema__)
        return crud



from pydantic import Field
from ardilla import Model


class _ForeignFieldMaker:
    NO_ACTION = 'NO ACTION'
    RESTRICT = 'RESTRICT'
    SET_NULL = 'SET NULL'
    SET_DEFAULT = 'SET DEFAULT'
    CASCADE = 'CASCADE'
    
    def __call__(
        self,
        *,
        references: Model,
        on_delete: str = NO_ACTION, 
        on_update: str = NO_ACTION,
    ):
        fk = getattr(references, '__pk__', None)
        tablename = getattr(references, '__tablename__', None)
        if not fk:
            raise ValueError('The referenced model requires to have a primary key')
        elif not tablename:
            raise ValueError('The referenced model does not have a tablename')
        return Field(
            references=tablename, 
            fk=fk,
            on_delete=on_delete,
            on_update=on_update,
        )

ForeignField = _ForeignFieldMaker()


"""
variables and functions here are used to generate and work with the Model's schemas
"""
import re
from typing import Optional
from datetime import datetime, date, time
from pydantic import BaseModel, Json
from .errors import ModelIntegrityError


SCHEMA_TEMPLATE: str = "CREATE TABLE IF NOT EXISTS {tablename} (\n{fields}\n);"


FIELD_MAPPING: dict[type, str] = {
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bool: "INTEGER",
    datetime: "DATETIME",
    bytes: "BLOB",
    date: "DATE",
    time: "TIME",
    Json: "TEXT",
}

AUTOFIELDS = {
    int: " AUTOINCREMENT",
    datetime: " DEFAULT CURRENT_TIMESTAMP",
    date: " DEFAULT CURRENT_DATE",
    time: " DEFAULT CURRENT_TIME"
}


def get_tablename(model: type[BaseModel]) -> str:
    """returns the tablename of a model either from the attribute __tablenam__
    or from the lowercase model's name

    Args:
        model (type[BaseModel]): the model

    Returns:
        str: the name of the table
    """
    return getattr(model, "__tablename__", model.__name__.lower())


def get_fields_schemas(Model: type[BaseModel]) -> list[str]:
    """Generates the fields for the table schema of the passed model

    Args:
        model (type[BaseModel]): the model

    Raises:
        TypeError: if the field is not listed in ardilla.schemas.FIELD_MAPPING
        ModelIntegrityError: if there's a Model field that is marked as pk but
            the model's __pk__ attr points to a different field

    Returns:
        str: a string containing the formatted fields to be added to a table schema
    """
    schemas = []
    constrains = []
    pk = None
    for field in Model.__fields__.values():
        name = field.name
        T = field.type_
        default = field.default
        extra = field.field_info.extra
        auto = extra.get('auto')
        autoerror = ModelIntegrityError(f'field {name} has a type of "{T}" which does not support "auto"')
        schema = f'{name} {FIELD_MAPPING[T]}'
        for k in {'pk', 'primary', 'primary_key'}:
            if k in extra and extra[k]:
                if pk is not None:
                    raise ModelIntegrityError('Only one primary key per model is allowed')
                elif hasattr(Model, '__pk__') and Model.__pk__ != name:
                    raise ModelIntegrityError(f"field {name} is marked as pk, but __pk__ points to another field.")
                
                pk = name
                schema += ' PRIMARY KEY'

                if auto and T in AUTOFIELDS:
                    schema += AUTOFIELDS[T]
                elif auto:
                    raise autoerror
                break
        else:
            if auto and T in AUTOFIELDS.keys() - {int}:
                schema += AUTOFIELDS[T]
            elif auto:
                raise autoerror
            elif default:
                if T in {int, str, float, bool}:
                    schema += f' DEFAULT {default!r}'
                elif T in {datetime, date, time}:
                    schema += f' DEFAULT {default}'
                elif T is bytes:
                    schema += f" DEFAULT (X'{default.hex()}')"
            elif field.required:
                    schema += ' NOT NULL'
                    
        schemas.append(schema)

        if extra.get('references'):
            references, fk, on_delete, on_update = (extra.get(f) for f in ['references', 'fk', 'on_delete', 'on_update'])
            constrains.append(
                f'FOREIGN KEY ({name}) '
                f'REFERENCES {references}({fk}) '
                f'ON UPDATE {on_update} '
                f'ON DELETE {on_delete}'
            )
            
    return schemas + constrains


def make_table_schema(Model: type[BaseModel]) -> str:
    """Generates the schema from a model based on its field configuration

    Args:
        Model (type[BaseModel]): the model

    Returns:
        str: the generated schema
    """
    tablename = get_tablename(Model)
    fields = get_fields_schemas(Model)
    schema = (
        f'CREATE TABLE IF NOT EXISTS {tablename}(\n' +
        ',\n'.join(f'\r    {f}' for f in fields) +
        '\n);'
    )
    return schema


def get_pk(schema: str) -> Optional[str]:
    """Gets the primary key field name from the passed schema

    Args:
        schema (str): table schema

    Returns:
        Optional[str]: the name of the primary key if any
    """
    # Check if the schema contains a primary key definition
    if "PRIMARY KEY" in schema:
        # Use a regular expression to extract the primary key column name
        match = re.search(r"(?i)\b(\w+)\b\s+(?:\w+\s+)*PRIMARY\s+KEY", schema)
        if match:
            return match.group(1)
    return None



import random
import sqlite3
from contextlib import contextmanager, asynccontextmanager
from pathlib import Path

import pytest

from pydantic import Field

from ardilla import Model
from ardilla.asyncio import Engine, Crud
from ardilla.errors import QueryExecutionError
from ardilla.fields import ForeignField


path = Path(__file__).parent
db = path / "testdb_async.sqlite"


class User(Model):
    id: int = Field(primary=True)
    name: str
    age: int


@asynccontextmanager
async def cleanup():
    try:
        db.unlink(missing_ok=True)
        engine = Engine(db)
        crud = engine.crud(User)
        yield crud
    finally:
        db.unlink(missing_ok=True)


@contextmanager
def query():
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        yield cur
    finally:
        cur.close()
        con.close()


@pytest.mark.asyncio
async def test_insert():
    async with cleanup() as crud:
        u = await crud.insert(id=1, name="chris", age=35)
        assert u is not None, "User wasn't created as expected"

        try:
            await crud.insert(id=1, name="chris", age=35)
        except QueryExecutionError:
            pass
        else:
            raise Exception("QueryExcecutionError should have been rised")


@pytest.mark.asyncio
async def test_insert_or_ignore():
    async with cleanup() as crud:
        u = await crud.insert_or_ignore(id=1, name="chris", age=35)
        assert u is not None, "User wasn't created as expected"
        u = await crud.insert_or_ignore(id=1, name="chris", age=35)
        assert u is None, "User was returned but it should have been None"


@pytest.mark.asyncio
async def test_save_one():
    async with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        await crud.save_one(chris)

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 1, "Incorrect number of users in the database"


@pytest.mark.asyncio
async def test_save_many():
    async with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        await crud.save_many(chris, moni, elena)

        with query() as cur:
            cur.execute("SELECT * FROM user;")
            res = cur.fetchall()

        assert len(res) == 3, "Incorrect number of users in the database"


@pytest.mark.asyncio
async def test_get_all():
    async with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        await crud.save_many(chris, moni, elena)

        users = await crud.get_all()
        assert len(users) == 3, "Incorrect number of users returned"


@pytest.mark.asyncio
async def test_get_many():
    async with cleanup() as crud:
        u1 = User(id=1, name="chris", age=35)
        u2 = User(id=2, name="chris", age=35)
        u3 = User(id=3, name="chris", age=35)
        u4 = User(id=4, name="moni", age=34)
        u5 = User(id=5, name="fran", age=1)
        await crud.save_many(u1, u2, u3, u4, u5)

        users = await crud.get_many(name="chris")

        assert len(users) == 3, "Incorrect number of users returned"

@pytest.mark.asyncio
async def test_get_many_with_limit():
    users = [User(id=n, name='chris', age=random.randint(1, 27)) for n in range(50)]
    async with cleanup() as crud:
        await crud.save_many(*users)
        users = await crud.get_many(limit=5)
        assert len(users) == 5

@pytest.mark.asyncio
async def test_get_many_with_ordering(): 
    async with cleanup() as crud:
        await crud.insert(id=1, name='chris', age=3)
        await crud.insert(id=2, name='chris', age=2)
        await crud.insert(id=3, name='chris', age=5)
        await crud.insert(id=4, name='chris', age=1)
        await crud.insert(id=5, name='chris', age=4)
        
        users = await crud.get_many(order_by={'age': 'desc'})
        
        sorted_users = sorted(users, key=lambda u: u.age, reverse=True)
        assert users == sorted_users

@pytest.mark.asyncio
async def test_get_or_create():
    async with cleanup() as crud:
        elena = User(id=1, name="elena", age=5)
        await crud.save_one(elena)

        u, c = await crud.get_or_create(name="fran", age=1)
        assert c is True, "User should have been created"
        assert u == User(id=2, name="fran", age=1), f"Unexpected user was created: {u}"
        u, c = await crud.get_or_create(id=1)
        assert u == elena, "Mismatch between expected and returned user"
        assert c is False, "User shouldn't have been created"


@pytest.mark.asyncio
async def test_get_or_none():
    async with cleanup() as crud:
        elena = User(id=1, name="elena", age=5)
        await crud.save_one(elena)

        u = await crud.get_or_none(id=1)
        assert u is not None, "User wasn't found but it should have"
        u = await crud.get_or_none(id=2)
        assert u is None, "User shouln't have been found but was"


@pytest.mark.asyncio
async def test_delete_one():
    async with cleanup() as crud:
        chris = User(id=1, name="chris", age=35)
        moni = User(id=2, name="moni", age=36)
        elena = User(id=3, name="elena", age=5)
        await crud.save_many(chris, moni, elena)
        await crud.delete_one(moni)
        users = await crud.get_all()
        assert len(users) == 2, "Delete one didn't delete the correct amount of users"

class Foo(Model):
    a: str
    b: int

@pytest.mark.asyncio
async def test_delete_one_without_ids():
    db.unlink(missing_ok=True)
    engine = Engine(db)
    crud = engine.crud(Foo)
    try:
        for i, l in enumerate('abcdef',1):
            await crud.insert(a=l, b=i)
            
        to_del = [
            Foo(a='b', b=2),
            Foo(a='c', b=3),
            Foo(a='d', b=4),
        ]
        for obj in to_del:
            await crud.delete_one(obj)
        
        foos = await crud.get_all()
        assert len(foos) == 3, 'Mismatch beween expected and found after delete_many'
        
    finally:
        db.unlink(missing_ok=True)


@pytest.mark.asyncio
async def test_delete_many_by_id():
    async with cleanup() as crud:
        users = [
            User(id=n, name='chris', age=n)
            for n in range(10)
        ]
        await crud.save_many(*users)
        
        to_delete = users[:-1]
        await crud.delete_many(*to_delete)
        
        users = await crud.get_all()
        assert len(users) == 1, "Delete many didn't delete the correct amount of users"

@pytest.mark.asyncio
async def test_delete_many_by_rowid():
    async with cleanup() as crud:
        users = [
            User(id=n, name='chris', age=n)
            for n in range(10)
        ]
        await crud.save_many(*users)
       
        await crud.delete_many(*users[:-1])
        
        users = await crud.get_all()
        
        
        assert len(users) == 1, "Delete many didn't delete the correct amount of users"

@pytest.mark.asyncio
async def test_foreign_keys():
    db = path / 'async_test.sqlite'
    db.unlink(missing_ok=True)
    engine = Engine(db, enable_foreing_keys=True)
    
    class Guild(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        
    class User(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        guild_id: int = ForeignField(references=Guild, on_delete=ForeignField.CASCADE)
    print(User.__schema__)
    
    gcrud = engine.crud(Guild)
    ucrud = engine.crud(User)
    
    ga = await gcrud.insert(name='guild a')
    gb = await gcrud.insert(name='guild b')
    for guild in [ga, gb]:
        for n in range(5):
            await ucrud.insert(name=f'user {n}', guild_id=guild.id)
    
    users = await ucrud.get_all()
    assert len(users) == 10
    await gcrud.delete_one(ga)
    users = await ucrud.get_all()
    assert len(users) == 5 
    db.unlink(missing_ok=True)


"""
Methods here are used by Crud classes to obtain the query 
strings and variable tuples to pass to the connections and cursors
"""
from typing import Any, Optional
from .errors import BadQueryError
from .models import M
from .ordering import validate_ordering
from .logging import log_query

Query = str
Vals = tuple[Any, ...]


def for_get_or_none(tablename: str, kws: dict) -> tuple[Query, Vals]:
    """called by _get_or_none_one method
    Args:
        tablename (str): name of the table
        kws (dict): the keywords to identify the rows with
    Returns:
        tuple[Query, Vals]: the query and values.  
    """
    keys, vals = zip(*kws.items())
    to_match = f" AND ".join(f"{k} = ?" for k in keys)
    q = f"SELECT rowid, * FROM {tablename} WHERE ({to_match}) LIMIT 1;"
    log_query(q, vals)
    return q, vals

def for_get_many(
        Model: M, *, 
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        kws: dict
    ) -> tuple[str, tuple[Any, ...]]:    
    """called by _get_many method
    Args:
     Args:
        Model (Model): the model of the table
        order_by (dict[str, str] | None ): 
            if passed Defines the sorting methods for the query
            defaults to no sorting
        limit (int | None) an integer to determine the number of items to grab
        kws (dict): the keywords to identify the rows with    
    """
    tablename = Model.__tablename__
    columns = tuple(Model.__fields__)
    
    if kws:
        keys, vals = zip(*kws.items())
        to_match = f" AND ".join(f"{k} = ?" for k in keys)
        filter_ = f' WHERE ({to_match})'
    else:
        filter_ = ''
        vals = ()
        
    if order_by is not None:
        ord = validate_ordering(columns, order_by)
        order_by = f' ORDER BY ' + ', '.join(f'{k} {v}' for k,v in ord.items())
    else:
        order_by = ''
    
    if limit is not None:
        if not isinstance(limit, int) or limit < 1: 
            raise ValueError('Limit, when passed, must be an integer larger than zero')
        limit_q = ' LIMIT ?'
        vals += limit,
    else:
        limit_q = ''

    
    q = f"SELECT rowid, * FROM {tablename}{filter_}{order_by}{limit_q};"
    return q, vals

def for_do_insert(
    tablename: str,
    ignore: bool,
    returning: bool,
    kws: dict,
) -> tuple[Query, Vals]:
    """called by _do_insert methods

    Args:
        tablename (str): name of the table
        ignore (bool): whether or not to use `INSERT OR IGNORE` vs just `INSERT`
        returning (bool): if the inserted values should be returned by the query
        kws (dict): the keywords representing column name and values

    Returns:
        tuple[Query, Vals]: the queries and values
    """
    keys, vals = zip(*kws.items())
    placeholders = ", ".join("?" * len(keys))
    cols = ", ".join(keys)

    q = "INSERT OR IGNORE " if ignore else "INSERT "
    q += f"INTO {tablename} ({cols}) VALUES ({placeholders})"
    q += " RETURNING *;" if returning else ";"
    log_query(q, vals)
    return q, vals


def for_save_one(obj: M) -> tuple[Query, Vals]:
    """called by save_one methods

    Args:
        obj (M): the Model instance to save

    Returns:
        tuple[Query, Vals]: the query and values
    """    
    cols, vals = zip(*obj.dict().items())

    if obj.__rowid__ is not None:
        q = f"""
        UPDATE {obj.__tablename__} SET {', '.join(f'{k} = ?' for k in cols)} WHERE rowid = ?;
        """
        vals += obj.__rowid__,

    else:
        placeholders = ", ".join("?" * len(cols))
        q = f"""
        INSERT OR REPLACE INTO {obj.__tablename__} ({', '.join(cols)}) VALUES ({placeholders});
        """
    log_query(q, vals)
    return q, vals

def for_save_many(objs: tuple[M]) -> tuple[Query, Vals]:
    """called by save_many methods

    Args:
        objs (tuple[M]): the objects to save

    Raises:
        BadQueryError: if the objs tuple is empty

    Returns:
        tuple[Query, Vals]: the query and values
    """
    if not objs:
        raise BadQueryError('To save many, you have to at least past one object')
    cols = tuple(objs[0].__fields__)
    tablename = objs[0].__tablename__
    placeholders = ", ".join("?" * len(cols))
    q = f'INSERT OR REPLACE INTO {tablename} ({", ".join(cols)}) VALUES ({placeholders});'
    vals = tuple(tuple(obj.dict().values()) for obj in objs)
    log_query(q, vals)
    return q, vals

def for_delete_one(obj: M) -> tuple[Query, Vals]:
    """called by delete_one methods

    Args:
        obj (M): the object to delete

    Returns:
        tuple[Query, Vals]: the query and values
    """    
    tablename = obj.__tablename__
    if obj.__pk__:
        q = f'DELETE FROM {tablename} WHERE {obj.__pk__} = ?'
        vals = getattr(obj, obj.__pk__),
    elif obj.__rowid__:
        q = f'DELETE FROM {tablename} WHERE rowid = ?'
        vals = obj.__rowid__,
    else:
        obj_dict = obj.dict()
        placeholders = " AND ".join(f"{k} = ?" for k in obj_dict)
        vals = tuple(obj_dict[k] for k in obj_dict)
        q = f"""
        DELETE FROM {tablename} WHERE ({placeholders});
        """
    log_query(q, vals)
    return q, vals

def for_delete_many(objs: tuple[M]) -> tuple[Query, Vals]:
    """called by delete_many methods

    Args:
        objs (tuple[M]): objects to delete

    Raises:
        IndexError: if the the obj tuple is empty
        BadQueryError: if the objects don't have either rowid or pks

    Returns:
        tuple[Query, Vals]: _description_
    """
    if not objs:
        raise IndexError('param "objs" is empty, pass at least one object')

    tablename = objs[0].__tablename__
    placeholders = ', '.join('?' for _ in objs)
    if all(obj.__rowid__ for obj in objs):
        vals = tuple(obj.__rowid__ for obj in objs)
        q = f'DELETE FROM {tablename} WHERE rowid IN ({placeholders})'

    elif pk := objs[0].__pk__:
        vals = tuple(getattr(obj, pk) for obj in objs)
        q = f'DELETE FROM {tablename} WHERE id IN ({placeholders})'
        
    else:
        raise BadQueryError('Objects requiere either a primary key or the rowid set for mass deletion')
    
    log_query(q, vals)
    return q, vals




from logging import getLogger
from typing import Optional

log = getLogger('ardilla')

def log_query(q: str, vals:  Optional[tuple] = None):
    vals = vals or ()
    log.debug(f'Querying: {q} - values: {vals}')




import sqlite3
from pathlib import Path
from datetime import datetime

from ardilla import Model, Field

from pydantic import Json


def test_default_tablename():
    class Foo(Model):
        id: int
    
    assert Foo.__tablename__ == "foo"

def test_field_pk():
    class Foo(Model):
        id: str = Field(primary=True)
    
    assert Foo.__pk__ == 'id'

def test_int_pk_auto():
    class Foo(Model):
        id: int = Field(pk=True, auto=True)
    
    schema = Foo.__schema__
    assert 'id INTEGER PRIMARY KEY AUTOINCREMENT' in schema



binary_data = b'some weird data'

class Complex(Model):
    id: int = Field(pk=True, auto=True)
    created: datetime = Field(auto=True)
    name: str = 'me'
    lastname: str | None = None
    foo: str
    data: bytes = binary_data


def test_default_schema():
    complex_schema = f'''
    \rCREATE TABLE IF NOT EXISTS complex(
    \r    id INTEGER PRIMARY KEY AUTOINCREMENT,
    \r    created DATETIME DEFAULT CURRENT_TIMESTAMP,
    \r    name TEXT DEFAULT 'me',
    \r    lastname TEXT,
    \r    foo TEXT NOT NULL,
    \r    data BLOB DEFAULT (X'{binary_data.hex()}')
    \r);
    '''
    assert Complex.__schema__.strip() == complex_schema.strip()


def test_complex_schema_works():
    try:
        db = Path(__file__).parent / 'db.sqlite3'
        db.unlink(missing_ok=True)
        con = sqlite3.connect(db)
        con.execute(Complex.__schema__)
        con.commit()
    finally:
        con.close()
        db.unlink(missing_ok=True)


class User(Model):
    id: int = Field(primary=True)
    name: str


tablename = "user"
schema = """
CREATE TABLE IF NOT EXISTS user(
\r    id INTEGER PRIMARY KEY,
\r    name TEXT NOT NULL
);
"""

def test_default_schema():

    assert User.__schema__.strip() == schema.strip()


def test_pk():
    assert User.__pk__ == "id"
